/* getopt - get option letter from argv                                     */

/* Copyright 1984  Michael M Rubenstein                                     */
/* Based on getopt.c from Henry Spencer @ U of Toronto Zoology              */

/* getopt(argc, argv, optstring)                                            */

/* optstring is a list of legal option characters.                          */
/* If a character is followed by a ":", the option takes a                  */
/* value, which may be separated by white space.  If a "#" appears in       */
/* optstring, a number is permitted with no option character.  ":" and "#"  */
/* may not be option characters.                                            */

/* Options which do not take values may be strung together ("-abc") or      */
/* separated ("-a -b -c") or mixed ("-ab -c").                              */

/* Returns option character, '?' (illegal option), '#' (number), or EOF     */
/* (no option found).  If illegal option, a message is printed on stderr.   */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

char                    *optarg;        /* Global argument pointer. */
int                     optind = 0;     /* Global argv index. */
int                     opterr = 1;     /* NZ to allow error messages */

static char             *scan = NULL;   /* Private scan pointer. */

extern char             *strchr();

int getopt(int argc, char *argv[], char *optstring)
{
  int                   c;
  char                  *place;

  optarg = NULL;
  if (optind == 0)
    scan = NULL;

  if (scan == NULL || *scan == '\0')
  {
    if (optind == 0)
      ++optind;

    if (optind >= argc || argv[optind][0] != '-')
      return EOF;

    if ((argv[optind][1] == '\0') != 0)
    {
      ++optind;
      return EOF;
    }
    
    scan = argv[optind++] + 1;
  }

  c = *(scan++);
  place = strchr(optstring, c);

  if (place == NULL && isdigit(c) && strchr(optstring, '#') != NULL)
  {
    --scan;
    c = '#';
  }
  else
  if (place == NULL || c == ':' || c == '#')
  {
    if (opterr)
      fprintf(stderr, "Unknown option - %c\n", c);
    return '?';
  }

  if (c == '#' || *(place + 1) == ':')
    if (*scan != '\0')
    {
      optarg = scan;
      scan = NULL;
    }
    else
      if (optind < argc)
        optarg = argv[optind++];
      else
        optarg = "";

  return c;
}

